'use strict';
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
exports.__esModule = true;
var retryCounter = Symbol('retryCounter');
// #region Task Queues
var TaskQueue = /** @class */ (function () {
    function TaskQueue(toleration) {
        var _this = this;
        if (toleration === void 0) { toleration = 3; }
        this.queue = [];
        this.errorStack = []; // errors occured while processing
        this._abortFlag = false; // is abort
        this._doneFlag = false; // has done
        this._handleFlag = false; // is handlering
        this.total = 0;
        this.succeed = 0;
        this.failed = 0;
        this.executed = 0;
        this.toleration = 0;
        this.__resolve = null;
        this.__reject = null;
        /**
         * Add a task to the queue
         * @param {TaskCapsule|function} task Task to be executed
         */
        this.add = function (task) {
            task[retryCounter] = 0;
            _this.queue.push(task);
        };
        this.__prepareToBegin = function () {
            if (!_this._handleFlag) {
                _this._abortFlag = false;
                _this._doneFlag = false;
                _this._handleFlag = true;
                _this.total = _this.queue.length;
                _this.succeed = 0;
                _this.failed = 0;
                _this.executed = 0;
                _this.errorStack = [];
                return true;
            }
            return false;
        };
        this.__execAmountChange = function (num) {
            _this.executed += num;
            if (_this.executed <= 0 && _this.queue.length === 0) {
                _this.__finish();
            }
        };
        this.__finish = function () {
            _this._handleFlag = false;
            _this._doneFlag = true;
            var flag = _this.failed === 0;
            if (flag) {
                _this.__resolve();
            }
            else {
                _this.__reject(_this.errorStack);
            }
        };
        this.toleration = toleration;
    }
    TaskQueue.prototype.__consumeValid = function () {
        if (this._doneFlag) {
            if (this.executed > 0) {
                console.warn('Called consume after the tasks were handled.');
            }
            return false;
        }
        else if (this._abortFlag) {
            this.__finish();
            return false;
        }
        else if (this.queue.length === 0) {
            return false;
        }
        else {
            return true;
        }
    };
    TaskQueue.prototype.abort = function () {
        this._abortFlag = true;
    };
    return TaskQueue;
}());
var ParallelQueue = /** @class */ (function (_super) {
    __extends(ParallelQueue, _super);
    function ParallelQueue(_a) {
        var _b = _a.limit, limit = _b === void 0 ? 5 : _b, _c = _a.span, span = _c === void 0 ? 300 : _c, toleration = _a.toleration;
        var _this = _super.call(this, toleration) || this;
        _this.__start = function () {
            if (_this.__consumeValid()) {
                if (_this.executed < _this.limitation && _this.queue.length > 0) {
                    var task_1 = _this.queue.shift();
                    _this.__execAmountChange(1);
                    task_1().then(function () {
                        // exec success
                        _this.succeed += 1;
                        _this.__execAmountChange(-1);
                    }, function (err) {
                        if (task_1[retryCounter] >= _this.toleration) {
                            // retried many times still failed
                            _this.failed += 1;
                            _this.errorStack.push(err);
                        }
                        else {
                            // failed but retry it
                            task_1[retryCounter] += 1;
                            _this.queue.unshift(task_1);
                        }
                        _this.__execAmountChange(-1);
                    });
                }
                if (_this.timespan > 0) {
                    setTimeout(_this.__start, _this.timespan);
                }
                else {
                    _this.__start();
                }
            }
        };
        _this.limitation = limit;
        _this.timespan = span;
        return _this;
    }
    ParallelQueue.prototype.consume = function () {
        var _this = this;
        return new Promise(function (resolve, reject) {
            _this.__resolve = resolve;
            _this.__reject = reject;
            if (_this.__prepareToBegin()) {
                if (_this.queue.length > 0) {
                    for (var _i = 0; _i < _this.limitation; _i += 1) {
                        _this.__start();
                    }
                }
                else {
                    _this.__execAmountChange(0);
                }
            }
            else {
                _this.__reject(new Error('Unable to consume'));
            }
        });
    };
    return ParallelQueue;
}(TaskQueue));
exports.ParallelQueue = ParallelQueue;
var SerialQueue = /** @class */ (function (_super) {
    __extends(SerialQueue, _super);
    function SerialQueue(_a) {
        var _b = _a.abortAfterFail, abortAfterFail = _b === void 0 ? false : _b, toleration = _a.toleration;
        var _this = _super.call(this, toleration) || this;
        _this.abortAfterFail = abortAfterFail;
        return _this;
    }
    SerialQueue.prototype.__start = function () {
        var _this = this;
        if (this.__consumeValid()) {
            this.__execAmountChange(1);
            var task_2 = this.queue.shift();
            task_2().then(function () {
                _this.succeed += 1;
                _this.__execAmountChange(-1);
                _this.__start();
            }, function (err) {
                if (task_2[retryCounter] >= _this.toleration) {
                    // retried many times still failed
                    _this.failed += 1;
                    _this.errorStack.push(err);
                    if (_this.abortAfterFail) {
                        _this.__finish();
                        return false;
                    }
                }
                else {
                    // failed but retry it
                    task_2[retryCounter] += 1;
                    _this.queue.unshift(task_2);
                }
                _this.__execAmountChange(-1);
                _this.__start();
            });
        }
    };
    SerialQueue.prototype.consume = function () {
        var _this = this;
        return new Promise(function (resolve, reject) {
            _this.__resolve = resolve;
            _this.__reject = reject;
            if (_this.__prepareToBegin()) {
                if (_this.queue.length > 0) {
                    _this.__start();
                }
                else {
                    _this.__execAmountChange(0);
                }
            }
            else {
                _this.__reject(new Error('Unable to consume'));
            }
        });
    };
    SerialQueue.prototype.abort = function () {
        this._abortFlag = true;
    };
    return SerialQueue;
}(TaskQueue));
exports.SerialQueue = SerialQueue;
//# sourceMappingURL=index.js.map